@@ -1,19 +1,26 @@
 import os
-import typing
 from datetime import datetime
 from importlib import import_module
 from typing import Dict, Callable, Optional, List, Tuple, Union
+from inspect import getfullargspec
 
-from twitchbot.database import CustomCommand
-from twitchbot.message import Message
-from twitchbot.database.dbcounter import increment_or_add_counter
+from .database import CustomCommand
+from .message import Message
+from .database.dbcounter import increment_or_add_counter
 from .config import cfg
 from .enums import CommandContext
-from .util import get_py_files, get_file_name
-from .util import temp_syspath
-
-if typing.TYPE_CHECKING:
-    from .modloader import Mod
+from .exceptions import InvalidArgumentsError
+from .translations import translate
+from .util import (
+    get_py_files,
+    get_file_name,
+    convert_args_to_function_parameter_types,
+    temp_syspath,
+    AutoCastFail,
+    AutoCastError,
+    get_callable_arg_types,
+    Param,
+)
 
 DEFAULT_COOLDOWN_BYPASS = 'bypass_cooldown'
 DEFAULT_COOLDOWN = 0
@@ -60,6 +67,9 @@ class Command:
         self.parent: Optional[Command] = None
         self.update_parent_command(parent)
 
+        if not self.syntax and self.func is not None:
+            self.syntax = self._generate_syntax_string()
+
         if global_command:
             commands[self.fullname] = self
 
@@ -130,9 +140,67 @@ class Command:
 
         return self.sub_cmds[args[0].lower()].get_sub_cmd(args[1:])
 
+    def _check_casted_args_for_auto_cast_fails(self, casted_args):
+        from .exceptions import InvalidArgumentsError
+        for arg in casted_args:
+            if not isinstance(arg, AutoCastFail):
+                continue
+
+            # handle fails caused by custom _handle_auto_cast() functions on classes
+            if isinstance(arg.exception, AutoCastError):
+                if arg.reason is not None and arg.exception.send_reason_to_chat:
+                    raise InvalidArgumentsError(reason=arg.exception.reason)
+                else:
+                    print(f'==== AUTO CAST FAIL ====\nreason = {arg.exception.reason}\n========================')
+
+                return True
+
+            if arg.param.annotation in (int, float):
+                raise InvalidArgumentsError(reason=translate('auto_cast_fail_number', arg_value=arg.value, arg_param_name=arg.param.name))
+
+        return False
+
+    def _check_args_fulfill_required_positional_arguments(self, args, function):
+        spec = getfullargspec(function)
+        # always subtract 1 because of msg parameter all commands have
+        # additionally, subtract an additional 1 if it's a mod command (has self/cls as its first parameter)
+        if spec.args and spec.args[0].casefold() in ('self', 'cls'):
+            offset = 2
+        else:
+            offset = 1
+
+        required_count = len(spec.args) - len(spec.defaults or ()) - offset
+
+        if len(args) < required_count:
+            raise InvalidArgumentsError(
+                reason=translate('args_does_not_fulfill_required_position_args', required_count=required_count, args_len=len(args)))
+
+    def _process_command_args_for_func(self, func, args):
+        casted_args = convert_args_to_function_parameter_types(func, args)
+        if self._check_casted_args_for_auto_cast_fails(casted_args):
+            return
+        self._check_args_fulfill_required_positional_arguments(args, func)
+        return casted_args
+
+    def _generate_syntax_string(self):
+        if self.func is None:
+            return ''
+
+        args = get_callable_arg_types(self.func, skip_self=True)[1:]
+        syntax_parts = []
+        for arg in args:
+            if arg.type == Param.VARARGS:
+                syntax_parts.append(f'({arg.name}...)')
+            elif arg.has_default_value:
+                syntax_parts.append(f'({arg.name}: {arg.default})')
+            else:
+                syntax_parts.append(f'<{arg.name}>')
+
+        return ' '.join(syntax_parts)
+
     async def execute(self, msg: Message):
         func, args = self._get_cmd_func(msg.parts[1:])
-        await func(msg, *args)
+        await func(msg, *self._process_command_args_for_func(func, args))
 
     async def has_permission_to_run_from_msg(self, origin_msg: Message):
         from .event_util import forward_event_with_results
@@ -156,6 +224,8 @@ class Command:
     # decorator support
     def __call__(self, func) -> 'Command':
         self.func = func
+        if not self.syntax:
+            self.syntax = self._generate_syntax_string()
         return self
 
     def __str__(self):
@@ -174,10 +244,6 @@ class SubCommand(Command):
         super().__init__(name=name, prefix='', func=func, permission=permission, syntax=syntax, help=help,
                          global_command=False, cooldown=cooldown, cooldown_bypass=cooldown_bypass, hidden=hidden, parent=parent)
 
-        # self.parent: Command = parent
-        # self.update_parent_command(parent)
-        # self.parent.sub_cmds[self.name] = self
-
 
 class DummyCommand(Command):
     def __init__(self, name: str, prefix: str = None, global_command: bool = True,
@@ -260,6 +326,7 @@ class ModCommand(Command):
 
     async def execute(self, msg: Message):
         func, args = self._get_cmd_func(msg.parts[1:])
+        args = self._process_command_args_for_func(func, args)
         if 'self' in func.__code__.co_varnames:
             await func(self.mod, msg, *args)
         else:
